/*
  -----------------------------------------------------------------------------
  - std-us-style.css                                                          -
  - AFNIX documentation style definition                                      -
  -----------------------------------------------------------------------------
  - This program is  free software;  you  can  redistribute it and/or  modify -
  - it provided that this copyright notice is kept intact.                    -
  -                                                                           -
  - This  program  is  distributed in the hope  that it  will be  useful, but -
  - without   any   warranty;  without  even   the   implied    warranty   of -
  - merchantability  or fitness for  a particular purpose. In not event shall -
  - the copyright holder be  liable  for  any direct, indirect, incidental or -
  - special damages arising in  any way out of the use of this software.      -
  -----------------------------------------------------------------------------
  - copyright (c) 1999-2023 amaury darsch                                     -
  -----------------------------------------------------------------------------
*/

/* ----------------------------------------------------------------------------
 * - body style                                                               -
 * ----------------------------------------------------------------------------
 */

body
{
    margin           : 20px 100px 20px 100px;
    padding          : 20px 0 0 0;
    font-family      : sans-serif; 
    font-size 	     : 10pt;
    font-style       : normal;
    font-weight      : normal;
    background-image : linear-gradient(#414963, #ffffff);
}

div.bbox
{
    background       : #ffffff;
    padding          : 40px;
}

/* ----------------------------------------------------------------------------
 * - title and sectioning                                                     -
 * ----------------------------------------------------------------------------
 */


div.chapter
{
    font-family      : sans-serif;
    font-size        : 12pt;
    font-style       : normal;
    font-weight      : bold;
    text-align       : left ;
    padding          : 3pt;
    color            : #414963;
    border-style     : none;
}

div.appendix
{
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : bold;
    text-align       : left ;
    padding          : 3pt;
    color            : #414963;
    border-style     : none;
}

div.section
{
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : bold;
    text-align       : left;
    padding          : 0pt;
    color            : #414963;
}

div.frame
{
    font-family      : monospace;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : left ;
    padding          : 6pt;
    color            : #777777;
    border-style     : solid;
    border-width     : 1px;
    border-color     : #dee7ec;
}

div.elem
{
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : bold;
    text-align       : left ;
    padding          : 3pt;
    background       : #414963;
    color            : #ffffff;
}

div.menu
{
    font-family      : sans-serif; 
    font-size        : 10pt;
    font-style 	     : normal;
    font-weight      : bold;
    text-align 	     : left;
    padding 	     : 3pt;
    background 	     : #dee7ec;
    color 	     : #000000;
}

div.text
{
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : left;
    padding          : 3pt;
    background       : #414963;
    color            : #ffffff;
}

div.head
{
    font-family      : sans-serif; 
    font-size 	     : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align 	     : right;
    padding 	     : 3pt;
    background 	     : #414963;
    color 	     : #ffffff;
}

@media print
{
    div.head
    {
	display	     : none;
	visibility   : none;
    }
}


pre.text
{
    font-family      : monospace;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : left;
    border-style     : solid;
    border-width     : 1px;
    border-color     : #8697cc;
    color            : #777777;
}

pre.frame
{
    font-family      : monospace;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : left;
    padding-left     : 6pt;
    color            : #777777;
    border-style     : solid;
    border-width     : 1px;
    border-color     : #8697cc;
}

pre.base
{
    font-family      : monospace;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : left;
    padding          : 3pt;
    margin-left      : 2em;
    color            : #777777;
    border-style     : solid;
    border-width     : 1px;
    border-color     : #8697cc;
}


/* ----------------------------------------------------------------------------
 * - base text                                                                -
 * ----------------------------------------------------------------------------
 */

p.title
{
    font-family      : sans-serif;
    font-size        : 18pt;
    font-style       : normal;
    font-weight      : bold;
    text-align       : right;
    color            : #8697cc;
}

p.subsect
{
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : bold;
    text-align       : left;
    color            : #414963;
}

p.text
{
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : justify;
    color            : black;
}

p.elem
{
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : bold;
    text-align       : left;
    color            : #8697cc;
}

h1.elem
{
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : left;
    color            : #black;
    padding-left     : 2em;
}

h1.text
{
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : justify;
    color            : #black;
    padding-left     : 2em;
}

h1.xref
{
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : left;
    color            : #8697cc;
}

h1.subt {
    border-left      : 2px solid #be1e2d;
    padding-left     : 10px;
    font-size        : 11pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : left;
    color            : #8697cc;
}

/* ----------------------------------------------------------------------------
 * - table formatting                                                         -
 * ----------------------------------------------------------------------------
 */

table
{
    width            : 100%;
    border-style     : none;
}

table.text
{
    width            : 100%;
    border-collapse  : collapse;
}


thead.text
{
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : bold;
    text-align       : left;
    color            : black;
}

tbody.text
{
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : left;
    color            : black;
}

th
{
    padding          : 3pt;
}

td
{
    padding          : 3pt;
}

td.logo
{
    font-family      : sans-serif; 
    font-size  	     : 18pt;
    font-style 	     : normal;
    font-weight      : bold;
    text-align 	     : left;
    vertical-align   : bottom;
    padding          : 0pt;
    color 	     : #414963;
}

td.title
{
    font-family      : sans-serif; 
    font-size  	     : 18pt;
    font-style 	     : normal;
    font-weight      : bold;
    text-align 	     : right;
    vertical-align   : bottom;
    padding          : 0pt;
    color 	     : #414963;
}

th.text
{
    border-style     : solid;
    border-width     : 1px;
    border-color     : #8697cc;
}

td.text
{
    border-style     : solid;
    border-width     : 1px;
    border-color     : #8697cc;
}

td.footer
{
    font-family      : sans-serif; 
    font-size        : 8pt;
    font-style       : normal;
    font-weight      : normal;
}

/* ----------------------------------------------------------------------------
 * - list formatting                                                          -
 * ----------------------------------------------------------------------------
 */

ul.text
{
    list-style-type  : square;
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : justify;
    color            : black;
}

ul.menu
{
    list-style-type  : square;
    font-family      : sans-serif;
    font-size        : 10pt;
    font-style       : normal;
    font-weight      : normal;
    text-align       : justify;
    padding	     : 3pt;
    margin 	     : 9pt;
    color            : #8697cc;
}

/* ----------------------------------------------------------------------------
 * - elements                                                                 -
 * ----------------------------------------------------------------------------
 */

img
{
    border-style     : none;
}

hr
{
    border-color     : #be1e2d;
    border-style     : solid;
    border-width     : 1px;
}

hr.title
{
    border-color     : #8697cc;
    border-style     : solid;
    border-width     : 5px;
}

em
{
    color            : #ff00ff;
}

i.product
{
    color            : #ff00ff;
}

i.package
{
    color            : #ff00ff;
}

i.item
{
    color            : #8697cc;
}

tt.command
{
    color            : #ff00ff;
}

tt.option
{
    color            : #ff00ff;
}

tt.file
{
    color            : #ff00ff;
}

tt.extn
{
    color            : #ff00ff;
}

tt.path
{
    color            : #ff00ff;
}

tt.code
{
    color            : #ff00ff;
}

a.chapter
{
    color            : black;
}

a.appendix
{
    color            : black;
}

a.client
{
    color            : black;
}

a:link
{
    text-decoration  : none;
}

a:visited
{
    text-decoration  : none;
}

a:active
{
    text-decoration  : none;
}

a:hover
{
    text-decoration  : underline;
}

/* ----------------------------------------------------------------------------
 * - print section                                                            -
 * ----------------------------------------------------------------------------
 */

@media print
{
    body
    {
	margin       : 20px 20px 20px 20px;
	padding      : 20px 0 0 0;
	font-family  : sans-serif; 
	font-size    : 10pt;
	font-style   : normal;
	font-weight  : normal;
    }
    
    div.bbox
    {
	background   : #ffffff;
	padding      : 10px;
    }
    
    div.head
    {
	display	     : none;
	visibility   : none;
    }
    colgroup#xref
    {
	display	     : none;
	visibility   : none;
    }
    td#xref
    {
	display	     : none;
	visibility   : none;
    }

    div.chapter
    {
        padding      : 0px;
    }

    div.section
    {
        padding      : 0px;
    }
}
